#include "devlock.h"

char pid_name[MAXNAMLEN];
char dev_name[MAXNAMLEN];
#ifdef MAJMIN
char pri_mm[MAXNAMLEN];
#endif	/* MAJMIN */
#ifdef HAS_CUA
char alt_name[MAXNAMLEN];
#ifdef MAJMIN
char alt_mm[MAXNAMLEN];
#endif	/* MAJMIN */
void __dumb_op( char *name );
#endif

/* Callable functions, part of library. */
pid_t devlock( char * );
int   devunlock( char * );
int   devrelock( char *, pid_t );

/* Internal functions */
int   __get_names( char * );
int   __write_lock( char *, pid_t );
pid_t __check_lock( char *, pid_t );

/* This is not the best implementation of the debugging
 * paradign. */
#ifdef MY_DEBUG
int debug;
#else
extern int debug;
#endif

/* 
 * This function creates a "LCK..<pid>" file and writes the current
 * pid to it in HDB format.  It then links the "LCK..<dev>" and the
 * "LCK..<alt>" to it.
 *
 * On success, a 0 is returned.
 *
 * If an active pid is found in an existing lockfile, that pid is
 * returned.  If the link fails for any other reason, all the
 * preceding successful links are unlinked and a -1 is returned.
 *
 * The "LCK..<pid>" name is unlinked before leaving whether
 * successful or failure.
 *
 * The argument, *dev, may be a full path name to the device driver
 * or just the name of the file.  That is, either "/dev/ttyS0" or
 * "ttyS0" will suffice.
 */

pid_t
devlock( char *dev )
{
	int old_mask;
	pid_t ret, pid, result;

	if( __get_names( dev ) < 0 ) {
		fprintf( stderr, "locks: Error in setting names.\n" );
		return( -1 );
	}
	if ( debug > 2 )
		fprintf( stderr, "locks: >%s< in >%s<\n", dev, dev_name );
	
	/*
	 * Set permissions to allow lockfiles to word read/write.
	 */
	old_mask = umask( 0 );
	ret = 0;
	pid = getpid();
	/* 
	 * This is fairly safe.  It is unlikely that our pid
	 * will have been recycled.
	 */
	if ( __write_lock( pid_name, pid ) < 0 ) {
		umask( old_mask );
		return( -1 );
	}
	umask( old_mask );
	if ( debug > 2 )
		fprintf( stderr, "lock 1: pid >%s<, dev >%s<, %d\n",
				 pid_name, dev_name, pid );
	
#ifdef MAJMIN
	/* 
	 * When used, pri_mm is the most generic so test for its
	 * existence first.  If there are collisions later, removing
	 * this will not cause problems.
	 */
	if (( result = link( pid_name, pri_mm )) < 0 ) {
		if(( ret = __check_lock( pri_mm, pid )) >= 0 )
			/* This is our own lockfile or a process exists */
			goto quit1;
		if ( debug > 2 )
			fprintf( stderr, "devlock: fail link pid to dev with pid %d\n",
					 ret );
	}
	if ( debug > 2 && result == 0 )
		fprintf( stderr, "lock 3: %s linked to %s\n", pid_name, pri_mm );
#endif	/* MAJMIN */
	if (( result = link( pid_name, dev_name )) < 0 ) {
		if(( ret = __check_lock( dev_name, pid )) >= 0 )
			/* This is our own lockfile or a process exists */
#ifdef MAJMIN
			goto quit2;
#else
			goto quit1;
#endif	/* MAJMIN */
		if ( debug > 2 )
			fprintf( stderr, "devlock: fail link pid to dev with pid %d\n",
					 ret );
	}
	if ( debug > 2 && result == 0 )
		fprintf( stderr, "lock 2: %s linked to %s\n", pid_name, dev_name );
#ifdef HAS_CUA
#ifdef MAJMIN
	if ( alt_mm[0] && (( result = link( pid_name, alt_mm )) < 0 )) {
		if(( ret = __check_lock( pri_mm, pid )) >= 0 )
			/* This is our own lockfile or a process exists */
			goto quit3;
		if ( debug > 2 )
			fprintf( stderr, "devlock: fail link pid to dev with pid %d\n",
					 ret );
	}
	if ( debug > 2 && result == 0 ) {
		fprintf( stderr, "lock 5: %s linked to %s\n", pid_name, alt_mm );
	}
#endif	/* MAJMIN */
	if ( alt_name[0] && (( result = link( pid_name, alt_name )) < 0 )) {
		if(( ret = __check_lock( alt_name, pid )) >= 0 ) {
			fprintf( stderr, "alt link identity\n" );
			/* This is our own lockfile or a process exists */
#ifdef MAJMIN
			goto quit4;
#else
			goto quit3;
#endif	/* MAJMIN */
		}
	}
	if ( debug > 2 && result == 0) {
		fprintf( stderr, "lock 4: %s linked to %s\n", pid_name, alt_name );
	}
#endif	/* HAS_CUA */
	/*
	 * Don't leave a stray file which will be orphaned
	 * if we crash before we can remove it.
	 */
	if( ret < 0 ) {
#ifdef HAS_CUA		
#ifdef MAJMIN
	quit4:
		unlink( alt_mm );
#endif	/* MAJMIN */
	quit3:
		unlink( dev_name );
#endif	/* HAS_CUA */
#ifdef MAJMIN
	quit2:
		unlink( pri_mm );
#endif	/* MAJMIN */
	}
 quit1:
	unlink( pid_name );
	return( ret );
}

int
devunlock( char *dev )
{
	if( __get_names( dev ) < 0 ) {
		fprintf( stderr, "unlock: Error in setting names.\n" );
		return( -1 );
	}
	if ( debug > 2 )
		fprintf( stderr, "Unlocking >%s< in >%s<\n", dev, dev_name );
	if( __check_lock( dev_name, 0 ) == 0 ) unlink( dev_name );
#ifdef MAJMIN
	if( __check_lock( pri_mm, 0 ) == 0 ) unlink( pri_mm );
#endif	/* MAJMIN */
#ifdef HAS_CUA
	if( alt_name[0] && (__check_lock( alt_name, 0 ) == 0 )) unlink( alt_name );
#ifdef MAJMIN
	if( __check_lock( alt_mm, 0 ) == 0 ) unlink( alt_mm );
#endif	/* MAJMIN */
#endif	/* HAS_CUA */
	return( 0 );
}

/* 
 * This is for use by a parent process which wants to hand off
 * the lockfiles to a child after forking.
 * 
 * It returns -1 on failure or 0 on success.
 */
int
devrelock( char *dev, pid_t chld_pid )
{
	pid_t my_pid, ret;

	if( __get_names( dev ) < 0 ) {
		fprintf( stderr, "relock: Error in setting names.\n" );
		return( -1 );
	}
	my_pid = getpid();
	if ( debug > 2 ) {
		fprintf( stderr, "relock: >%s< in >%s<\n", dev, dev_name );
		fprintf( stderr, "\tmy %d, chld %d\n", my_pid, chld_pid );
	}
	if (( ret = __check_lock( dev_name, my_pid )) == 0 )
		__write_lock( dev_name, chld_pid );
	else
		return( -1 );
#ifdef HAS_CUA
	/*
	 * If 'alt_name' is hard linked to 'dev_name', this step
	 * is redunant.  If we can make this assumption, this
	 * step can be deleted.
	 */
	if ( alt_name[0] ) {
		if (( ret = __check_lock( alt_name, my_pid )) == 0 )
			__write_lock( alt_name, chld_pid );
		else
			if( ret != chld_pid ) return( -1 );
	}
#endif	/* HAS_CUA */
	/* 
	 * We don't need to check the Major/Minor files as they are linked
	 * to the dev_name file.
	 */
	return( 0 );
}

/*
 * Open a file and write our pid into it.
 */
int
__write_lock( char *dev, pid_t my_pid )
{
	FILE *lock;

	if (( lock = fopen( dev, "w" )) == NULL ) {
		fprintf( stderr, "lock: Unable to open %s\t%s\n", dev,
				sys_errlist[errno] );
		return( -1 );
	}
	fprintf( lock, "%10d\n", my_pid );
	fclose( lock );
	return( 0 );
}

/*
 * Read the pid stored in an existing lockfile.  Then determine if
 * it our own or the child pid if relocking.
 * 
 * Return:
 *    -1 - in case of failure of any step.
 *     0 - it is our own or the process is dead.
 *   fpid - the pid found in the lockfile.
 */
pid_t
__check_lock( char *file, pid_t wrk_pid )
{
	char file_pid[12];
	int fd;
	pid_t fpid;

	if (( fd = open( file, O_RDONLY )) < 0 ) {
		/* Shut up extraneous messages. */
		if ( debug > 2 )
			fprintf( stderr, "check: Unable to open %s\t%s\n", file,
					 sys_errlist[errno] );
		return( -1 );
	}
	if ( read( fd, &file_pid, sizeof( file_pid ))
		 != sizeof( file_pid ) -1 ) {
		/* Shut up extraneous messages. */
		if ( debug > 2 )
			fprintf( stderr, "check: Unable to read from %s, %s\n",
					 file, sys_errlist[errno] );
		return( -1 );
	}
	close( fd );
	/*
	 * If this is a child pid, we have already rewritten
	 * a linked lockfile.
	 */
	file_pid[10] = '\0';
	if ( debug > 2 )
		fprintf(stderr, "lock %s, file %s, wrk %d, get %d\n",
				file, file_pid, wrk_pid, getpid());
	fpid = (pid_t)atoi( file_pid );
	if( fpid == wrk_pid )
		return( 0 );
		
	if ( fpid != getpid()) {
		errno = 0;
		
		if ( kill( fpid, 0 ) == 0 ) {
			/*
			 * This would be a nice message if the 'getty' didn't
			 * repeat it so often.
			 */
			if ( debug > 2 )
				fprintf( stderr, "check: %s locked by pid %d\t%s\n",
						 file, fpid, sys_errlist[errno] );
			return ( fpid );
		}
		/* The process doesn't exist */
		unlink( file );
		link( pid_name, file );
	}
	return( 0 );
}

/*
 * Generate the names for the lockfiles
 * These will be of two to five types.
 *
 * 1. /path/LCK..<pid>
 * 2. /path/LCK..<dev>
 * 3. /path/LCK..<maj>.<min>  For primary device
 * 4. /path/LCK..<alt>	(Used If old cua* devices are used)
 * 5. /path/LCK..<maj>.<min>  For alternate device
 *
 * The <pid> is removed before leaving the calling function.
 *
 * These names will be placed in the global char arrays:
 *    pid_name[]
 *    dev_name[]
 *    pri_mm[]    if MAJMIN is defined
 *    alt_name[]  if HAS_CUA is defined
 *    alt_mm[]    if HAS_CUA and MAJMIN are defined
 *    
 * Return:
 *   -1 - A failure occurred in some step.
 *    0 - Success
 */
int
__get_names( char *dev )
{
	DIR *dev_dir;
	struct stat mystat;
	struct dirent *dir_p;
	int node, is_sym;
	char *p, *q, *s;
	char lock_name[MAXNAMLEN];

	/*
	 * Mark the names as undefined.
	 */
	pid_name[0] = '\0';
	dev_name[0] = '\0';
#ifdef MAJMIN
	pri_mm[0] = '\0';
#endif
#ifdef HAS_CUA
	alt_name[0] = '\0';
#ifdef MAJMIN
	alt_mm[0] = '\0';
#endif
#endif	/* HAS_CUA */

	/* 
	 * Scan past the path portion of 'dev' if it was included in
	 * 'dev'.
	 */
	p = dev;
	q = DEV_PATH;
	s = p + MAXNAMLEN - 1;
	while ( *p == *q && p < s ) p++, q++;

	/*
	 * Step past a trailing '/'.
	 */
	if( !*q && *p == '/' ) p++;

	/*
	 * Have we gone too far?
	 */
	if ( p >= s || !*p ) {
		fprintf( stderr, "get_names: The the device name is faulty. '%s'\n",
				 dev );
		return( -1 );
	}

	/* 
	 * Now attach to the path.
	 */
	strncpy( lock_name, p, MAXNAMLEN - 1 );
	lock_name[MAXNAMLEN] = '\0';
	sprintf( dev_name, "%s%s", DEV_PATH, lock_name );

	is_sym = 0;
	if ( lstat( dev_name, &mystat) < 0 ) {
		if ( debug > 2 )
			fprintf( stderr, "get_names: Unable to lstat %s\n\t%s\n",
					 dev_name, sys_errlist[errno] );
	}
	else is_sym = S_ISLNK( mystat.st_mode );

	if ( stat( dev_name, &mystat ) < 0 ) {
		if ( debug > 2 )
			fprintf( stderr, "get_names: Unable to stat %s\n\t%s\n",
					 dev_name, sys_errlist[errno] );
		return( -1 );
	}

	/*
	 * Reuse this array
	 */
	dev_name[0] = '\0';

	/* 
	 * Create the "LCK..<pid>" file name
	 */
	sprintf( pid_name, "%sLCK..%d", LOCK_PATH, getpid() );

#ifdef MAJMIN
	/* 
	 * Create the "LCK..<maj>.<min>" file name
	 */
	sprintf( pri_mm, "%sLCK..%d.%d", LOCK_PATH, MAJOR( mystat.st_rdev ),
			MINOR( mystat.st_rdev ));
#endif	/* MAJMIN */

	/*
	 * Open the '/dev' directory so that we can search for the
	 * name of the device.  This also works if the 'dev' name
	 * given to us was a symbolic link.
	 */
	node = mystat.st_ino;
	if( is_sym ) {
		if(( dev_dir = opendir( DEV_PATH )) == NULL ) {
			fprintf( stderr, "get_names: Unable to opendir %s\n\t%s\n",
					DEV_PATH, sys_errlist[errno] );
		}
		else {
			while(( dir_p = readdir( dev_dir )) != NULL ) {
				if( dir_p->d_ino == node ) {
					strncpy( lock_name, dir_p->d_name, MAXNAMLEN - 1 );
					lock_name[MAXNAMLEN] = '\0';
					break;
				}
			}
			closedir( dev_dir );
		}
	}
	sprintf( dev_name, "%sLCK..%s", LOCK_PATH, lock_name );
#ifdef HAS_CUA
	__dumb_op( lock_name );
#endif	/* HAS_CUA */
	return( 0 );
}

#ifdef HAS_CUA
void
__dumb_op( char *name )
{
	char *p, *q;
#ifdef MAJMIN
	char wrk_name[MAXNAMLEN];
	struct stat mystat;
#endif	/* MAJMIN */

	p = name;
	q = TTYS;
	while( *p == *q ) p++, q++;
	if( *q != '\0' ) {
		p = name;
		q = CUA;
		while( *p == *q ) p++, q++;
		if( *q != '\0' ) {
			fprintf( stderr, "dumb_op: No compatible device name found.\n" );
			return;
		}
		q = TTYS;
	}
	else q = CUA;
	sprintf( alt_name, "%sLCK..%s%s", LOCK_PATH, q, p );

#ifdef MAJMIN
	sprintf( wrk_name, "%s%s%s", DEV_PATH, q, p );
	if ( stat( wrk_name, &mystat ) < 0 ) {
		if ( debug > 2 )
			fprintf( stderr, "dumb_op: Unable to stat %s\n\t%s\n",
					 wrk_name, sys_errlist[errno] );
		return;
	}
	/* 
	 * Create the "LCK..<maj>.<min>" file name
	 */
	sprintf( alt_mm, "%sLCK..%d.%d", LOCK_PATH, MAJOR( mystat.st_rdev ),
			MINOR( mystat.st_rdev ));
	if ( debug > 2 )
		fprintf( stderr, "HAS_CUA: alt_mm %s\n", alt_mm );
#endif	/* MAJMIN */
}
#endif	/* HAS_CUA */
